// reverse_scriptor.cpp : Defines the entry point for the console application.
// By BS-er, with a little help from BigBadBogie

#include "stdafx.h"
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <string>
#include <set>

#include <AtlBase.h>
#include <AtlConv.h>

using namespace std;

struct StepData
{
	short Instruction; // The switch-case branch that this step will use.
	short Item1; // These depend on the function used for this step.  They can be
	short Item2; // indices to handles, vectors, strings etc., or for other information.
	short Item3;
	short Item4;
};

struct ObjectiveStruct
{
	char s[150];
	inline const char& operator[](int n) const	{ return s[n]; }
	inline operator const char*() const { return s; }
};

struct NameStruct
{
	char s[40];
	inline char& operator[](int n) { return s[n]; }
	inline operator const char*() const { return s; }
	inline operator char*() const { return (char*)s; }
};

struct Vector
{
	float x;
	float y;
	float z;
};

short NumberOfRoutines = 0;

char* Image;

// Pointers to arrays of data in the DLL
StepData* RoutineData;
Vector* GlobalVectorArray;
NameStruct* NameArray;
short* GlobalRoutineSpeed;
short* GlobalIntegerArray;
bool* GlobalRoutinePriority;
short* RoutineOffset;
ObjectiveStruct* ObjectiveArray;
bool* DebugText;

int MaxObject;
int MaxVector;
int MaxVariable;
int MaxText;
static char TextBuffer[256];

static bool ObjectReferenced[96];

typedef set<short> set_short;

namespace Reverse { string OperationToText(const StepData &Input); }

// ----------------------------------------------------------------------------
bool CheckAboutText(const char* Text, const char* CompareText)
{
	bool Result = true;
	for (int i=0; CompareText[i]; i++)
	{
		if (Text[i] != CompareText[i]) { Result = false; break; }
	}
	return Result;
}

// ----------------------------------------------------------------------------
int _tmain(int argc, _TCHAR* argv[])
{
//	int dummy;
//	scanf("%i", &dummy); // This helps if you need to attach to the process for debugging.

	// Loop through all the file paths passed in.
	for (int i=1; i<argc; i++)
	{
		USES_CONVERSION;
		for (int j=0; j<96; j++) { ObjectReferenced[j] = false; }
		string FilePath = W2A(argv[i]);
		printf((FilePath + '\n').c_str());
		set_short LabelSet;
		string RoutineText[30];
		MaxObject = -1;
		MaxVector = -1;
		MaxVariable = -1;
		MaxText = -1;

		ifstream File(argv[i], ios_base::binary | ios_base::in);
		if (!File.good() || File.eof() || !File.is_open()) { continue; }
		size_t DotPos = FilePath.find_last_of('.');
		if (DotPos == string::npos) { continue; }
		if (DotPos >= FilePath.size() - 3) { continue; }
		FilePath[DotPos + 1] = 'b'; FilePath[DotPos + 2] = 'z'; FilePath[DotPos + 3] = 's';
		File.seekg(0, ios_base::beg);
		ifstream::pos_type begin_pos = File.tellg();
		File.seekg(0, ios_base::end);
		int FileSize = static_cast<int>(File.tellg() - begin_pos);
		File.seekg(0, ios_base::beg);
		bool Valid = false;
		Image = 0;

		// Use the file size to determine which scriptor DLL image this came from
		// (not the most reliable or certain means, but it will do).
		switch(FileSize)
		{
		case 98304:
			Image = new char[98304];
			File.read(Image, 98304);
			// Check for DLL Scriptor signature text
			if (CheckAboutText(Image + 0x149cc, "DLL created"))
			{
				RoutineData = (StepData*)(Image + 0x10040);
				NumberOfRoutines = *(short*)(Image + 0x12750);
				GlobalVectorArray = (Vector*)(Image + 0x12758);
				NameArray = (NameStruct*)(Image + 0x128d8);
				GlobalRoutineSpeed = (short*)(Image + 0x14048);
				GlobalIntegerArray = (short*)(Image + 0x14070);
				GlobalRoutinePriority = (bool*)(Image + 0x140b0);
				RoutineOffset = (short*)(Image + 0x140c4);
				ObjectiveArray = (ObjectiveStruct*)(Image + 0x140ec);
				DebugText = (bool*)(Image + 0x149b8);
				Valid = true;
			}
			break;
		case 118784:
			Image = new char[118784];
			File.read(Image, 118784);
			// Check for DLL Scriptor signature text
			if (CheckAboutText(Image + 0x17838, "DLL created"))
			{
				RoutineData = (StepData*)(Image + 0x11040);
				NumberOfRoutines = *(short*)(Image + 0x146f0);
				GlobalVectorArray = (Vector*)(Image + 0x146f8);
				NameArray = (NameStruct*)(Image + 0x14938);
				GlobalRoutineSpeed = (short*)(Image + 0x16878);
				GlobalIntegerArray = (short*)(Image + 0x168b4);
				GlobalRoutinePriority = (bool*)(Image + 0x16914);
				RoutineOffset = (short*)(Image + 0x16934);
				ObjectiveArray = (ObjectiveStruct*)(Image + 0x16970);
				DebugText = (bool*)(Image + 0x17818);
				Valid = true;
			}
			break;
		case 110592:
			Image = new char[110592];
			File.read(Image, 110592);
			// Check for DLL Scriptor signature text
			if (CheckAboutText(Image + 0x159cc, "DLL created"))
			{
				RoutineData = (StepData*)(Image + 0x11040);
				NumberOfRoutines = *(short*)(Image + 0x13750);
				GlobalVectorArray = (Vector*)(Image + 0x13758);
				NameArray = (NameStruct*)(Image + 0x138d8);
				GlobalRoutineSpeed = (short*)(Image + 0x15048);
				GlobalIntegerArray = (short*)(Image + 0x15070);
				GlobalRoutinePriority = (bool*)(Image + 0x150b0);
				RoutineOffset = (short*)(Image + 0x150c4);
				ObjectiveArray = (ObjectiveStruct*)(Image + 0x150ec);
				DebugText = (bool*)(Image + 0x159b8);
				Valid = true;
			}
			else if (CheckAboutText(Image + 0x178e2, "DLL Scriptor"))
			{
				RoutineData = (StepData*)(Image + 0x10040);
				NumberOfRoutines = *(short*)(Image + 0x136F0);
				GlobalVectorArray = (Vector*)(Image + 0x136F8);
				NameArray = (NameStruct*)(Image + 0x13938);
				GlobalRoutineSpeed = (short*)(Image + 0x15878);
				GlobalIntegerArray = (short*)(Image + 0x158B4);
				GlobalRoutinePriority = (bool*)(Image + 0x15914);
				RoutineOffset = (short*)(Image + 0x15934);
				ObjectiveArray = (ObjectiveStruct*)(Image + 0x15970);
				DebugText = (bool*)(Image + 0x16818);
				Valid = true;
			}
			break;
		case 122880:
			Image = new char[122880];
			File.read(Image, 122880);
			// Check for DLL Scriptor signature text
			if (CheckAboutText(Image + 0x1997e, "DLL Scriptor"))
			{
				RoutineData = (StepData*)(Image + 0x12040);
				NumberOfRoutines = *(short*)(Image + 0x156F0);
				GlobalVectorArray = (Vector*)(Image + 0x156F8);
				NameArray = (NameStruct*)(Image + 0x15938);
				GlobalRoutineSpeed = (short*)(Image + 0x17878);
				GlobalIntegerArray = (short*)(Image + 0x178B4);
				GlobalRoutinePriority = (bool*)(Image + 0x17914);
				RoutineOffset = (short*)(Image + 0x17934);
				ObjectiveArray = (ObjectiveStruct*)(Image + 0x17970);
				DebugText = (bool*)(Image + 0x18818);
				Valid = true;
			}
			break;
		case 69632:
			Image = new char[69632];
			File.read(Image, 69632);
			// Check for DLL Scriptor signature text
			if (CheckAboutText(Image + 0xe870, "ScriptorVM2"))
			{
				RoutineData = (StepData*)(Image + 0x8030);
				NumberOfRoutines = *(short*)(Image + 0xB6E0);
				GlobalVectorArray = (Vector*)(Image + 0xB6E8);
				NameArray = (NameStruct*)(Image + 0xB928);
				GlobalRoutineSpeed = (short*)(Image + 0xD868);
				GlobalIntegerArray = (short*)(Image + 0xD8A4);
				GlobalRoutinePriority = (bool*)(Image + 0xD904);
				RoutineOffset = (short*)(Image + 0xD924);
				ObjectiveArray = (ObjectiveStruct*)(Image + 0xD960);
				DebugText = (bool*)(Image + 0xE808);
				Valid = true;
			}
			break;
		}
		if (Valid)
		{
			// Process the routines for offset location labels
			for (int i=0; i<NumberOfRoutines; i++)
			{
				short CurrentOffset = RoutineOffset[i];
				// Loop through to find the used label offsets for jumping and branching
				while (RoutineData[CurrentOffset].Instruction != 0)
				{
					// Switch-case for those operations that use offset location labels
					switch(RoutineData[CurrentOffset].Instruction)
					{
					case 27:
						LabelSet.insert(RoutineData[CurrentOffset].Item1);
						break;
					case 10: case 11: case 12: case 13: case 31: case 165: case 166:
					case 167: case 168:case 211: case 212: case 213: case 214:
						LabelSet.insert(RoutineData[CurrentOffset].Item2);
						break;
					case 87:
						LabelSet.insert(RoutineData[CurrentOffset].Item1);
						if (RoutineData[CurrentOffset].Item2 >= 0)
						{
							LabelSet.insert(RoutineData[CurrentOffset].Item2);
						}
						if (RoutineData[CurrentOffset].Item3 >= 0)
						{
							LabelSet.insert(RoutineData[CurrentOffset].Item3);
						}
						if (RoutineData[CurrentOffset].Item4 >= 0)
						{
							LabelSet.insert(RoutineData[CurrentOffset].Item4);
						}
						break;
					}
					CurrentOffset++;
				}
			}
			
			set_short::iterator LabelIter = LabelSet.begin();

			// Process the routines to assemble the text
			for (int i=0; i<NumberOfRoutines; i++)
			{
				RoutineText[i] = "//--------------------------------------------------------------------\r\n";
				sprintf_s(TextBuffer, 256, "[routine,_Routine%i,%i,", i+1, GlobalRoutineSpeed[i]);
				RoutineText[i] += TextBuffer;
				if (GlobalRoutinePriority[i]) { RoutineText[i] += "true"; } else { RoutineText[i] += "false"; }
				if (DebugText[i]) { RoutineText[i] += ",true"; }
				RoutineText[i] += "]\r\n";
				short CurrentOffset = RoutineOffset[i];
				// Loop through to assemble routine text
				while (true)
				{
					if (LabelIter != LabelSet.end())
					{
						// Check to see if the next label applies to this offset
						if (*LabelIter == CurrentOffset)
						{
							sprintf_s(TextBuffer, 256, "LOC_%i:\r\n",CurrentOffset);
							LabelIter++;
							RoutineText[i] += TextBuffer;
						}
					}
					if (RoutineData[CurrentOffset].Instruction == 0) { break; }
					RoutineText[i] += '\t' + Reverse::OperationToText(RoutineData[CurrentOffset]) + "\r\n";
					CurrentOffset++;
				}
			}

			// Open the output file and put the header information
			ofstream OutFile(FilePath.c_str());
			if (OutFile.is_open())
			{
				OutFile << "\r\n[objects]\r\n" << endl;
				for (int i=0; i<=MaxObject; i++)
				{
					int Count = 1;
					for (int j=i+1; j<=MaxObject; j++)
					{
						if (ObjectReferenced[j])
						{
							break;
						}
						else
						{
							Count++;
						}
					}
					if (Count > 1)
					{
						sprintf_s(TextBuffer, 256, "Object%i <%i>", i + 1, Count);
						i += (Count - 1);
					}
					else
					{
						sprintf_s(TextBuffer, 256, "Object%i", i + 1);
					}
					OutFile << TextBuffer << endl;
				}
				OutFile << "\r\n[positions]\r\n" << endl;
				for (int i=0; i<=MaxVector; i++)
				{
					sprintf_s(TextBuffer, 256, "Position%i, %g,%g,%g", i + 1, GlobalVectorArray[i].x,
						GlobalVectorArray[i].y, GlobalVectorArray[i].z);
					OutFile << TextBuffer << endl;
				}
				OutFile << "\r\n[variables]\r\n" << endl;
				for (int i=0; i<=MaxVariable; i++)
				{
					int Value = GlobalIntegerArray[i];
					sprintf_s(TextBuffer, 256, "Variable%i, %i", i + 1, Value);
					OutFile << TextBuffer << endl;
				}
				OutFile << "\r\n[text]\r\n" << endl;
				for (int i=0; i<=MaxText; i++)
				{
					string Text = (const char*)ObjectiveArray[i];
					size_t FindPos = 0;
					// Convert carriage returns to their symbolic form.
					while (true)
					{
						size_t NextPos = Text.find_first_of(10, FindPos);
						if (NextPos == string::npos) { break; }
						Text[NextPos] = '\\';
						Text.insert(NextPos + 1, "n");
						FindPos = NextPos + 1;
					}
					sprintf_s(TextBuffer, 256, "_Text%i,\r\n", i + 1);
					OutFile << TextBuffer << '\"' << Text << '\"' << endl << endl;
				}

				// Put the routines
				for (int i=0; i<NumberOfRoutines; i++)
				{
					OutFile << RoutineText[i] << endl;
				}
			}
		}
		delete[] Image;
	}
	return 0;
}

namespace Reverse
{


static enum Operand_Types
{
    BLANK,
    OBJECT,
    STRING,
    TEXT,
    VARIABLE,
    ROUTINE,
    POSITION,
    LOCATION,
    INTEGER,
    BOOLEAN,
    COLOR,
    OPTIONAL_POSITION,
    OPTIONAL_INTEGER,
    OPTIONAL_BOOLEAN,
    OPTIONAL_LOCATION,
    OPTIONAL_VARIABLE,
    INTEGER_OR_BOOLEAN,
    DISABLE_OR_INTEGER,
	OPTIONAL_OBJECT
};

static const char* ColorNames[] =
{
	"white", "green", "red", "turquoise", "yellow", "blue", "violet"
};


// ----------------------------------------------------------------------------
// This function converts the type and argument value into text.
// ----------------------------------------------------------------------------
string ArgToString(Operand_Types Type, short Arg)
{
	TextBuffer[0] = 0;
	switch (Type)
	{
	case OBJECT: case OPTIONAL_OBJECT:
		if (Arg > 95) { sprintf_s(TextBuffer, 256, ",ERROR: Object index out of range"); break; }
		if (Arg >= 0) { sprintf_s(TextBuffer, 256, ",Object%i", Arg + 1); }
		if (Arg > MaxObject) { MaxObject = Arg; }
		ObjectReferenced[Arg] = true;
		break;
	case STRING:
		sprintf_s(TextBuffer, 256, ",\"%s\"", (const char*)NameArray[Arg]); break;
	case TEXT:
		sprintf_s(TextBuffer, 256, ",_Text%i", Arg + 1);
		if (Arg > MaxText) { MaxText = Arg; } break;
	case VARIABLE: case OPTIONAL_VARIABLE:
		if (Arg >= 0) { sprintf_s(TextBuffer, 256, ",Variable%i", Arg + 1); }
		if (Arg > MaxVariable) {MaxVariable = Arg; }
		break;
	case ROUTINE:
		sprintf_s(TextBuffer, ",_Routine%i", Arg + 1); break;
	case POSITION: case OPTIONAL_POSITION:
		if (Arg >= 0) { sprintf_s(TextBuffer, 256, ",Position%i", Arg + 1); }
		if (Arg > MaxVector) { MaxVector = Arg; }
		break;
	case INTEGER: case INTEGER_OR_BOOLEAN: case DISABLE_OR_INTEGER:
		sprintf_s(TextBuffer, 256, ",%i", Arg); break;
	case OPTIONAL_INTEGER:
		if (Arg != 0) { sprintf_s(TextBuffer, 256, ",%i", Arg); } break;
	case BOOLEAN:
		if (Arg != 0) { sprintf_s(TextBuffer, 256, ",%s", "true"); }
		else { sprintf_s(TextBuffer, ",%s", "false"); } break;
	case OPTIONAL_BOOLEAN:
		if (Arg != 0) { sprintf_s(TextBuffer, 256, ",%s", "true"); } break;
	case COLOR:
		if ((Arg >= 0) && (Arg <= 6))
		{
			sprintf_s(TextBuffer, ",%s", ColorNames[Arg]);
		}
		break;
	case LOCATION: case OPTIONAL_LOCATION:
		if (Arg >= 0) { sprintf_s(TextBuffer, 256, ",LOC_%i", Arg); } break;
		break;
	}
	
	return TextBuffer;
}

// Ugly but time-saving macro for converting the operation data to a string
#define OPDATA(Id, Name, arg1, arg2, arg3, arg4) case Id: \
	if ((arg1 == DISABLE_OR_INTEGER) && (Input.Item1 == 0)) \
	{ Result = Name + ArgToString(arg1, Input.Item1); } \
	else { Result = Name + ArgToString(arg1, Input.Item1) + ArgToString(arg2, Input.Item2) + \
		ArgToString(arg3, Input.Item3) + ArgToString(arg4, Input.Item4); } \
	break;


// ----------------------------------------------------------------------------
string OperationToText(const StepData &Input)
{
	string Result;
	switch (Input.Instruction)
	{
	default: // Fall through to case 0.
	case 0:
		break;
	OPDATA ( 60, "AddAmmo",        OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 59, "AddHealth",      OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 95, "AddScrap",       INTEGER, INTEGER, BLANK, BLANK)
	OPDATA (  8, "Add",            VARIABLE, INTEGER, OPTIONAL_VARIABLE, BLANK)
	OPDATA (117, "AddToPos",       INTEGER, INTEGER, INTEGER, POSITION)
	OPDATA ( 49, "AllLookAt",      INTEGER, OBJECT, INTEGER, BLANK)
	OPDATA ( 69, "Ally",           INTEGER, INTEGER, BLANK, BLANK)
	OPDATA (  4, "Attack",         OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA ( 32, "Audio",          STRING, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA ( 33, "BeaconOn",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 34, "BeaconOff",      OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 26, "Build",          OBJECT, STRING, INTEGER, OPTIONAL_VARIABLE)
	OPDATA ( 89, "CamObject",      OBJECT, POSITION, OBJECT, INTEGER)
	OPDATA ( 92, "CamPos",         OBJECT, POSITION, POSITION, INTEGER)
	OPDATA ( 90, "CamPath",        STRING, INTEGER, INTEGER, OBJECT)
	OPDATA ( 91, "CamPathDir",     STRING, INTEGER, INTEGER, INTEGER_OR_BOOLEAN)
	OPDATA ( 40, "ChangeSide",      BLANK, BLANK, BLANK, BLANK)
	OPDATA (  5, "Clear",          BLANK, BLANK, BLANK, BLANK )
	OPDATA (101, "ClearTeamColor", INTEGER, BLANK, BLANK, BLANK)
	OPDATA ( 84, "ConsoleResult",  BLANK, BLANK, BLANK, BLANK)
	OPDATA ( 77, "ConsoleWord",    STRING, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (  1, "Create",         OBJECT, STRING, INTEGER, POSITION)
	OPDATA ( 25, "Createp",        OBJECT, STRING, INTEGER, STRING)
	OPDATA ( 28, "Defend",         OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA (  6, "Display",        TEXT, COLOR, OPTIONAL_VARIABLE, OPTIONAL_VARIABLE)
	OPDATA ( 21, "DistObject",     OBJECT, OBJECT, OPTIONAL_VARIABLE, BLANK)
	OPDATA ( 22, "DistPath",       OBJECT, STRING, OPTIONAL_VARIABLE, BLANK)
	OPDATA ( 47, "DropOff",        OBJECT, STRING, INTEGER, OPTIONAL_VARIABLE)
	OPDATA ( 81, "EjectPilot",     OBJECT, BLANK, BLANK, BLANK)
	OPDATA (114, "EvalPos",        POSITION, INTEGER, OPTIONAL_VARIABLE, BLANK)
	OPDATA (116, "Fade",           INTEGER, INTEGER, INTEGER, BLANK)
	OPDATA ( 18, "Fail",           INTEGER, STRING, BLANK, BLANK)
	OPDATA (  3, "Follow",         OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA (105, "GetAlt",         POSITION, OPTIONAL_POSITION, BLANK, BLANK)
	OPDATA (127, "GetByIndex",     OBJECT, OBJECT, VARIABLE, BLANK)
	OPDATA ( 41, "GetByLabel",     OBJECT, STRING, OPTIONAL_VARIABLE, BLANK)
	OPDATA (  7, "GetByTS",        OBJECT, INTEGER, INTEGER, BLANK)
	OPDATA ( 14, "GetCommand",     OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (111, "GetCompForce",   OPTIONAL_VARIABLE, BLANK, BLANK, BLANK)
	OPDATA ( 55, "GetCurAmmo",     OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA ( 53, "GetCurHealth",   OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA ( 57, "GetMaxHealth",   OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA ( 98, "GetMaxScrap",    INTEGER, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (112, "GetMySide",      OPTIONAL_VARIABLE, BLANK, BLANK, BLANK)
	OPDATA ( 45, "GetIn",          OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA (113, "GetMode",        OPTIONAL_VARIABLE, BLANK, BLANK, BLANK)
	OPDATA (110, "GetMyForce",     OPTIONAL_VARIABLE, BLANK, BLANK, BLANK)
	OPDATA ( 23, "GetPlayer",      OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 50, "GetPos",         OBJECT, POSITION, BLANK, BLANK)
	OPDATA ( 97, "GetScrap",       INTEGER, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (126, "GetRace",        OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (106, "GetTarget",      OBJECT, OBJECT, BLANK, BLANK)
	OPDATA ( 61, "GetTeam",        OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA ( 99, "GetTime",        OPTIONAL_VARIABLE, BLANK, BLANK, BLANK)
	OPDATA ( 72, "GetTimerTime",   OPTIONAL_VARIABLE, BLANK, BLANK, BLANK)
	OPDATA ( 93, "GetTug",         OBJECT, OBJECT, BLANK, BLANK)
	OPDATA ( 80, "GiveWeapon",     OBJECT, STRING, OPTIONAL_BOOLEAN, BLANK)
	OPDATA ( 85, "GlobalSpeed",    INTEGER, BLANK, BLANK, BLANK)
	OPDATA ( 16, "GoTo",           OBJECT, STRING, INTEGER, OPTIONAL_VARIABLE)
	OPDATA (102, "GoToo",          OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA ( 94, "HasCargo",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 51, "HasPilot",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 52, "HealthPercent",  OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA ( 82, "HopOut",         OBJECT, BLANK, BLANK, BLANK)
	OPDATA (108, "HoppedOutOf",    OBJECT, OBJECT, BLANK, BLANK)
	OPDATA ( 12, "IfEQ",           INTEGER_OR_BOOLEAN, LOCATION, BLANK, BLANK)
	OPDATA ( 11, "IfGT",           INTEGER, LOCATION, BLANK, BLANK)
	OPDATA ( 10, "IfLT",           INTEGER, LOCATION, BLANK, BLANK)
	OPDATA ( 13, "IfNE",           INTEGER_OR_BOOLEAN, LOCATION, BLANK, BLANK)
	OPDATA (109, "IsAlly",         OBJECT, OBJECT, BLANK, BLANK)
	OPDATA ( 17, "IsAround",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 78, "IsDeployed",     OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 29, "IsODF",          OBJECT, STRING, OPTIONAL_VARIABLE, BLANK)
	OPDATA (104, "IsPerson",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 79, "IsSelected",     OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 27, "JumpTo",         LOCATION, BLANK, BLANK, BLANK)
	OPDATA (107, "KillPilot",      OBJECT, BLANK, BLANK, BLANK)
	OPDATA ( 48, "LookAt",         OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA ( 43, "Mine",           OBJECT, STRING, INTEGER, BLANK)
	OPDATA (115, "Mult",           VARIABLE, INTEGER, OPTIONAL_VARIABLE, BLANK)
	OPDATA ( 35, "NearObject",     OBJECT, OBJECT, BLANK, BLANK)
	OPDATA ( 36, "NearVehicle",    OBJECT, OBJECT, BLANK, BLANK)
	OPDATA ( 37, "NearVehiclep",   OBJECT, STRING, INTEGER, OPTIONAL_VARIABLE)
	OPDATA ( 38, "NearBuilding",   OBJECT, OBJECT, BLANK, BLANK)
	OPDATA ( 39, "NearEnemy",      OBJECT, OBJECT, BLANK, BLANK)
	OPDATA (121, "OnDelObject",    DISABLE_OR_INTEGER, ROUTINE, OBJECT, BLANK)
	OPDATA (120, "OnNewObject",    DISABLE_OR_INTEGER, ROUTINE, OBJECT, BLANK)
	OPDATA ( 44, "Patrol",         OBJECT, STRING, INTEGER, OPTIONAL_VARIABLE)
	OPDATA ( 46, "Pickup",         OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA ( 64, "PlaceAt",        OBJECT, POSITION, BLANK, BLANK)
	OPDATA ( 63, "PlaceAtP",       OBJECT, STRING, OPTIONAL_VARIABLE, BLANK)
	OPDATA (119, "PosAngle",       POSITION, VARIABLE, VARIABLE, POSITION)
	OPDATA ( 74, "Quake",          INTEGER, BLANK, BLANK, BLANK)
	OPDATA ( 87, "RandJump",       LOCATION, OPTIONAL_LOCATION, OPTIONAL_LOCATION, OPTIONAL_LOCATION)
	OPDATA ( 86, "RandPos",        POSITION, POSITION, POSITION, BLANK)
	OPDATA (118, "RandShift",      POSITION, INTEGER, INTEGER, POSITION)
	OPDATA (103, "Remove",         OBJECT, BLANK, BLANK, BLANK)
	OPDATA (  9, "RunSpeed",       ROUTINE, INTEGER, BOOLEAN, OPTIONAL_BOOLEAN)
	OPDATA ( 42, "Service",        OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA ( 15, "Set",            VARIABLE, INTEGER, BLANK, BLANK)
	OPDATA (128, "SetByIndex",     OBJECT, VARIABLE, OBJECT, BLANK)
	OPDATA ( 56, "SetCurAmmo",     OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 54, "SetCurHealth",   OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 88, "SetPos",         POSITION, INTEGER, INTEGER, INTEGER)
	OPDATA ( 58, "SetMaxHealth",   OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 24, "SetGroup",       OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 83, "SetName",        OBJECT, STRING, OPTIONAL_VARIABLE, BLANK)
	OPDATA ( 68, "SetPlan",        STRING, INTEGER, OPTIONAL_VARIABLE, BLANK)
	OPDATA ( 96, "SetScrap",       INTEGER, INTEGER, BLANK, BLANK)
	OPDATA (124, "SetSkill",       OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 31, "SetStep",        ROUTINE, LOCATION, BLANK, BLANK)
	OPDATA ( 67, "SetPTeam",       OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 62, "SetTeam",        OBJECT, INTEGER, BLANK, BLANK)
	OPDATA (125, "SetWeapon",      OBJECT, INTEGER, BLANK, BLANK)
	OPDATA ( 66, "ShotBy",         OBJECT, OBJECT, BLANK, BLANK)
	OPDATA (122, "StartSound",     STRING, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA ( 71, "StartTimer",     INTEGER, OPTIONAL_INTEGER, OPTIONAL_INTEGER, BLANK)
	OPDATA ( 20, "Stop",           OBJECT, INTEGER, BLANK, BLANK)
	OPDATA (123, "StopSound",      VARIABLE, BLANK, BLANK, BLANK)
	OPDATA ( 76, "StopQuake",      BLANK, BLANK, BLANK, BLANK)
	OPDATA ( 73, "StopTimer",      BLANK, BLANK, BLANK, BLANK)
	OPDATA ( 30, "StoreResult",    VARIABLE, BLANK, BLANK, BLANK)
	OPDATA ( 19, "Succeed",        INTEGER, STRING, BLANK, BLANK)
	OPDATA (100, "TeamColor",      INTEGER, INTEGER, INTEGER, INTEGER)
	OPDATA ( 70, "UnAlly",         INTEGER, INTEGER, BLANK, BLANK)
	OPDATA ( 65, "UnitSpeed",      OBJECT, INTEGER, INTEGER, INTEGER)
	OPDATA ( 75, "UpdateQuake",    INTEGER, BLANK, BLANK, BLANK)
	OPDATA (  2, "Wait",           INTEGER, BLANK, BLANK, BLANK)
	// New with 1.2:
	OPDATA (129, "CamOf",          OBJECT, INTEGER, INTEGER, OPTIONAL_POSITION)
	OPDATA (130, "IsSameObject",   OBJECT, OBJECT, BLANK, BLANK)
	OPDATA (131, "IsPlayer",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA (132, "IFaceExec",      STRING, BLANK, BLANK, BLANK)
	OPDATA (133, "IFaceActivate",  STRING, BLANK, BLANK, BLANK)
	OPDATA (134, "IFaceDeactivate",STRING, BLANK, BLANK, BLANK)
	OPDATA (135, "IFaceCommand",   STRING, BLANK, BLANK, BLANK)
	OPDATA (136, "IFaceString",    STRING, STRING, BLANK, BLANK)
	OPDATA (137, "IFaceSetString", STRING, STRING, BLANK, BLANK)
	OPDATA (138, "IFaceGetString", STRING, STRING, INTEGER, BLANK)
	OPDATA (139, "IFaceInt",       STRING, INTEGER, BLANK, BLANK)
	OPDATA (140, "IFaceSetInt",    STRING, INTEGER, BLANK, BLANK)
	OPDATA (141, "IFaceGetInt",    STRING, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (142, "IFaceClearList", STRING, BLANK, BLANK, BLANK)
	OPDATA (143, "IFaceAddText",   STRING, STRING, BLANK, BLANK)
	OPDATA (144, "IFaceGetSelect", STRING, STRING, INTEGER, BLANK)
	OPDATA (145, "PlayMovie",      STRING, BLANK, BLANK, BLANK)
	OPDATA (146, "IsFlying",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA (147, "Damage",         OBJECT, INTEGER, BLANK, BLANK)
	OPDATA (148, "IsInfo",         STRING, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (149, "StartTimerUp",   INTEGER, OPTIONAL_INTEGER, OPTIONAL_INTEGER, BLANK)
	OPDATA (150, "SetAsPlayer",    OBJECT, BLANK, BLANK, BLANK)
	OPDATA (151, "SetUserTarget",  OBJECT, BLANK, BLANK, BLANK)
	OPDATA (152, "GetUserTarget",  OBJECT, BLANK, BLANK, BLANK)
	OPDATA (153, "PlaybackVehicle",STRING, BLANK, BLANK, BLANK)
	OPDATA (154, "AddPilot",       OBJECT, BLANK, BLANK, BLANK)
	OPDATA (155, "SpawnBirds",     INTEGER, STRING, INTEGER, STRING)
	OPDATA (156, "RemoveBirds",    BLANK, BLANK, BLANK, BLANK)
	OPDATA (157, "SetAnimation",   OBJECT, STRING, INTEGER, BLANK)
	OPDATA (158, "StartAnimation", OBJECT, BLANK, BLANK, BLANK)
	// New for community project
	OPDATA (159, "FreeCamera",     OPTIONAL_BOOLEAN, BLANK, BLANK, BLANK)
	OPDATA (160, "FreeFinish",     OPTIONAL_BOOLEAN, BLANK, BLANK, BLANK)
	OPDATA (161, "SetPortal",      OBJECT, BLANK, BLANK, BLANK)
	OPDATA (162, "GetPortalDest",  OBJECT, OBJECT, BLANK, BLANK)
	OPDATA (163, "OnPortalDist",   DISABLE_OR_INTEGER, ROUTINE, OBJECT, INTEGER)
	OPDATA (164, "Teleport",       OBJECT, OBJECT, INTEGER, BLANK)
	OPDATA (183, "SetPortalDest",  OBJECT, OBJECT, BLANK, BLANK)
	OPDATA (173, "Move",           OBJECT, VARIABLE, VARIABLE, OPTIONAL_POSITION)
	OPDATA (177, "SetAngle",       OBJECT, INTEGER, BLANK, BLANK)
	OPDATA (165, "IfLTV",          VARIABLE, LOCATION, BLANK, BLANK)
	OPDATA (166, "IfGTV",          VARIABLE, LOCATION, BLANK, BLANK)
	OPDATA (167, "IfEQV",          VARIABLE, LOCATION, BLANK, BLANK)
	OPDATA (168, "IfNEV",          VARIABLE, LOCATION, BLANK, BLANK)
	OPDATA (169, "Independence",   OBJECT, INTEGER, BLANK, BLANK)
	OPDATA (170, "NearObjectCount",OBJECT, VARIABLE, INTEGER, STRING)
	OPDATA (171, "AddV",           VARIABLE, VARIABLE, OPTIONAL_VARIABLE, BLANK)
	OPDATA (172, "MultV",          VARIABLE, VARIABLE, OPTIONAL_VARIABLE, BLANK)
	OPDATA (174, "SetTeamV",       OBJECT, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (175, "Replace",        OBJECT, STRING, INTEGER_OR_BOOLEAN, BLANK)
	OPDATA (176, "CameraCut",      INTEGER_OR_BOOLEAN, BLANK, BLANK, BLANK)
	OPDATA (178, "UpdateQuakeV",   VARIABLE, BLANK, BLANK, BLANK)
	OPDATA (179, "AddHealthV",     OBJECT, VARIABLE, BLANK, BLANK)
	OPDATA (180, "Divide",         VARIABLE, INTEGER, OPTIONAL_VARIABLE, BLANK)
	OPDATA (181, "AngleObject",    OBJECT, OBJECT, OPTIONAL_VARIABLE, BLANK)
	OPDATA (182, "PosMultAdd",	   POSITION, POSITION, INTEGER, OPTIONAL_POSITION)
	OPDATA (184, "Retreat",        OBJECT, STRING, INTEGER, BLANK)
	OPDATA (185, "SetMaxAmmo",     OBJECT, INTEGER, BLANK, BLANK)
	OPDATA (186, "Deploy",         OBJECT, BLANK, BLANK, BLANK)
	OPDATA (187, "SetCurAmmoV",    OBJECT, VARIABLE, BLANK, BLANK)
	OPDATA (188, "SetCurHealthV",  OBJECT, VARIABLE, BLANK, BLANK)
	OPDATA (189, "GetEmptyGroup",  OPTIONAL_VARIABLE, BLANK, BLANK, BLANK)
	OPDATA (190, "PreloadAudio",   STRING, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (191, "IFaceSetIntV",   STRING, VARIABLE, BLANK, BLANK)
	OPDATA (192, "SetGroupV",      OBJECT, VARIABLE, BLANK, BLANK)
	OPDATA (193, "SetMaxHealthV",  OBJECT, VARIABLE, BLANK, BLANK)
	OPDATA (194, "SetMaxAmmoV",    OBJECT, VARIABLE, BLANK, BLANK)
	OPDATA (195, "HideTimer",      BLANK, BLANK, BLANK, BLANK)
	OPDATA (196, "AtTerminal",     OBJECT, OBJECT, BLANK, BLANK)
	OPDATA (197, "InBuilding",     OBJECT, OBJECT, BLANK, BLANK)
	OPDATA (198, "SetDisplayTime", INTEGER, BLANK, BLANK, BLANK)
	OPDATA (199, "DisplayFile",    STRING, COLOR, OPTIONAL_VARIABLE, OPTIONAL_VARIABLE)
	OPDATA (200, "CreateV",         OBJECT, STRING, VARIABLE, POSITION)
	OPDATA (201, "CreatepV",        OBJECT, STRING, VARIABLE, STRING)
	OPDATA (202, "NearObjectCountV",OBJECT, VARIABLE, VARIABLE, STRING)
	OPDATA (203, "SetPlanV",        STRING, VARIABLE, OPTIONAL_VARIABLE, BLANK)
	OPDATA (204, "SetScrapV",       VARIABLE, INTEGER, BLANK, BLANK)
	OPDATA (205, "AddScrapV",       VARIABLE, INTEGER, BLANK, BLANK)
	OPDATA (206, "GetScrapV",       VARIABLE, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (207, "GetMaxScrapV",    VARIABLE, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (208, "GetByTSV",        OBJECT, VARIABLE, INTEGER, BLANK)
	OPDATA (209, "WaitV",           VARIABLE, BLANK, BLANK, BLANK)
	OPDATA (210, "LeaveScrap",      OBJECT, BOOLEAN, BLANK, BLANK)
	OPDATA (211, "IfLE",            INTEGER, LOCATION, BLANK, BLANK)
	OPDATA (212, "IfGE",            INTEGER, LOCATION, BLANK, BLANK)
	OPDATA (213, "IfLEV",           VARIABLE, LOCATION, BLANK, BLANK)
	OPDATA (214, "IfGEV",           VARIABLE, LOCATION, BLANK, BLANK)
	// New with BZ 1.3 but will work in 1.2:
	OPDATA (215, "OnKilled",        DISABLE_OR_INTEGER, ROUTINE, OBJECT, OPTIONAL_OBJECT)
	OPDATA (216, "OnSniped",        DISABLE_OR_INTEGER, ROUTINE, OBJECT, OPTIONAL_OBJECT)
	OPDATA (217, "DllHandleDeath",  BLANK, BLANK, BLANK, BLANK)
	// New for BZ 1.3:
	OPDATA (218, "IsOneThree",      BLANK, BLANK, BLANK, BLANK)
	OPDATA (219, "ConsoleCommand",  STRING, OPTIONAL_VARIABLE, OPTIONAL_VARIABLE, OPTIONAL_VARIABLE)
	OPDATA (220, "IsWithinODF",		OBJECT, STRING, INTEGER, OPTIONAL_BOOLEAN)
	OPDATA (221, "DoesODFExist",    STRING, OPTIONAL_VARIABLE, BLANK, BLANK)
	OPDATA (222, "IsWeaponODF",     OBJECT, STRING, INTEGER, BLANK)
	// Will work in 1.2:
	OPDATA (223, "Halt",            BLANK, BLANK, BLANK, BLANK)
	OPDATA (224, "Swap",            OBJECT, OBJECT, BOOLEAN, BLANK)
	}
	return Result;
}


} // namespace Reverse